/* SMTP proxy in C language, based on Java version
 *    Jamis Buck <jgb3@email.byu.edu>
 *
 *    Copying policy: GPL
 */

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <netdb.h>
#include <stdarg.h>
#include <ctype.h>

#define MAX_ADDRESS_LENGTH       (  256 )
#define MAX_HELO_STRING_LENGTH   (  128 )
#define MAX_LOG_FILE_NAME_LENGTH (  128 )
#define MAX_BUFFER_LENGTH        ( 1024 )
#define MAX_HEADER_ITEM_LEN      (  512 )
#define SOCKET_ERROR             (   -1 )

#if defined( DEBUG )
# define DBG_MSG(x)  printf x
#else
# define DBG_MSG(x)  (void*)0
#endif

#define VERBOSE( parms, level, cmd ) \
  if( (parms)->verbose_level >= (level) ) { cmd; }

struct __system_parameters__ {
  int local_port;
  int remote_port;
  int max_clients;
  int verbose_level;
  char remote_address[ MAX_ADDRESS_LENGTH+1 ];
  char helo_string[ MAX_HELO_STRING_LENGTH+1 ];
  FILE* log_file;
};

typedef struct __system_parameters__ SYSTEM_PARAMETERS;

static int g_active_clients;


static void child_signal( int sig ) {
  pid_t pid;
  int   status;

  pid = wait( &status );
  DBG_MSG(( "child %d exited with a status of %d\n", pid, status ));
  signal( SIGCHLD, &child_signal );

  /* one less active client */
  g_active_clients--;
}


static void log_message( SYSTEM_PARAMETERS* parms, char* fmt, ... ) {
  va_list arg;
  FILE*   output;

  output = stderr;
  if( parms->log_file ) {
    output = parms->log_file;
  }

  va_start( arg, fmt );
  vfprintf( output, fmt, arg );
  va_end( arg );
  fputs( "\n", output );
  fflush( output );
}


static void report_error( char* msg, SYSTEM_PARAMETERS* parms ) {
  log_message( parms, "%s: error #%d (%s)\n", msg, errno, strerror( errno ) );
}


static void usage() {
  puts( "smtpproxy v. 0.1" );
  puts( "  by Jamis Buck (jgb3@email.byu.edu)" );
  puts( "    http://www.jamisbuck.org" );
  puts( "  based on the Java smtpproxy code by Radim Kolar, hsn@cybermail.net" );
  puts( "    http://home.worldonline.cz/~cz210552/" );
  puts( "" );
  puts( "  mandatory options:" );
  puts( "    -d <remote-address>" );
  puts( "       this is the address of the remote SMTP server or proxy" );
  puts( "" );
  puts( "  optional options:" );
  puts( "    -p <local-port> [default: 25]" );
  puts( "       this is the local port to bind to" );
  puts( "    -r <remote-port> [default: 25]" );
  puts( "       this is the port that the remote SMTP server is running on" );
  puts( "    -s <helo-string> [default: host's name]" );
  puts( "       this is the HELO string to use" );
  puts( "    -m <max-clients> [default: 5]" );
  puts( "       this is the maximum number of requests that may be serviced simultaneously" );
  puts( "" );
  puts( "  miscellaneous options:" );
  puts( "    -l <log-file>" );
  puts( "       log messages to the given file" );
  puts( "    -v" );
  puts( "       display status messages.  You may specify this option multiple times to" );
  puts( "       have more messages shown." );
  puts( "    -h" );
  puts( "       displays this help screen." );
  puts( "" );
}


static int read_line( int sock, char* buffer, int length, int *eof, SYSTEM_PARAMETERS* parms ) {
  char c;
  int  pos;

  pos = 0;
  *eof = 1;
  while( recv( sock, &c, 1, 0 ) == 1 ) {
    *eof = 0;
    buffer[ pos++ ] = c;
    if( c == '\r' ) {
      if( recv( sock, &c, 1, MSG_PEEK ) != 1 ) {
        report_error( "read_line -- \\r", parms );
        return -1;
      }
      if( c == '\n' ) {
        recv( sock, &c, 1, 0 );
        pos--;
      }
    }
    if( c == '\n' ) {
      break;
    }
    if( pos >= length-1 ) {
      break;
    }
  }

  buffer[ pos ] = 0;
  return pos;
}


static int read_reply( int input, int output, SYSTEM_PARAMETERS* parms ) {
  char line[ MAX_BUFFER_LENGTH ];
  int  eof;
  
  do {
    if( read_line( input, line, sizeof( line ), &eof, parms ) < 0 ) {
      report_error( "read_reply->read_line", parms );
      return -1;
    }

    strcat( line, "\r\n" );
    if( send( output, line, strlen(line), 0 ) < 0 ) {
      report_error( "read_reply->send 1", parms );
      return -1;
    }
  } while( strlen( line ) > 3 && line[3] == '-' );

  return 0;
}


static int smtp_server_connect( SYSTEM_PARAMETERS* parms ) {
  struct sockaddr_in addr;
  struct hostent*    phe;
  int                remote_server;

  DBG_MSG(( "smtp_server_connect -- initializing addr struct\n" ));
  memset( &addr, 0, sizeof( addr ) );
  addr.sin_family = AF_INET;
  addr.sin_port = htons( parms->remote_port );

  DBG_MSG(( "smtp_server_connect -- getting host by name of remote address (%s)\n", parms->remote_address ));
  if( phe = gethostbyname( parms->remote_address ) ) {
    memcpy( &addr.sin_addr, phe->h_addr, phe->h_length );
  } else {
    report_error( "gethostbyname", parms );
    return -1;
  }

  DBG_MSG(( "smtp_server_connect -- creating socket for remote connection\n" ));
  if( ( remote_server = socket( AF_INET, SOCK_STREAM, 0 ) ) == SOCKET_ERROR ) {
    report_error( "smtp_server_connect socket", parms );
    return -1;
  }

  DBG_MSG(( "smtp_server_connect -- connecting\n" ));
  if( connect( remote_server, (struct sockaddr*)&addr, sizeof( addr ) ) == SOCKET_ERROR ) {
    report_error( "smtp_server_connect connect", parms );
    close( remote_server );
    return -1;
  }

  DBG_MSG(( "smtp_server_connect -- success!\n" ));
  return remote_server;
}


static int server_socket( SYSTEM_PARAMETERS* parms ) {
  int                socket_handle;
  int                optval;
  struct sockaddr_in server;

  DBG_MSG(( "creating the server socket\n" ));
  if( ( socket_handle = socket( PF_INET, SOCK_STREAM, 0 ) ) == SOCKET_ERROR ) {
    report_error( "create socket", parms );
    return -1;
  }

  DBG_MSG(( "configuring socket (%d)\n", socket_handle ));
  memset( &server, 0, sizeof( server ) );
  server.sin_family      = AF_INET;
  server.sin_port        = htons( parms->local_port );
  server.sin_addr.s_addr = htonl( INADDR_ANY );
	
  DBG_MSG(( "making address reusable\n" ));
  if( setsockopt( socket_handle, SOL_SOCKET, SO_REUSEADDR, (char *)&optval, sizeof(optval) ) == SOCKET_ERROR ) {
    close( socket_handle );
    return -1;
  }

  DBG_MSG(( "binding to port %d\n", parms->local_port ));
  if( bind( socket_handle, (struct sockaddr*)&server, sizeof( server ) ) == SOCKET_ERROR ) {
    report_error( "bind socket", parms );
    close( socket_handle );
    return -1;
  }

  DBG_MSG(( "beginning listen\n" ));
  if( listen( socket_handle, parms->max_clients ) == SOCKET_ERROR ) {
    report_error( "listen", parms );
    close( socket_handle );
    return -1;
  }

  DBG_MSG(( "server_socket completed successfully (socket=%d)\n", socket_handle ));
  return socket_handle;
}


static int server_accept( int server, SYSTEM_PARAMETERS* parms ) {
  int                client;
  int                client_addr_len;
  struct sockaddr_in client_addr;

  client_addr_len = sizeof( client_addr );
  while( 1 ) {
    if( ( client = accept( server, (struct sockaddr*)&client_addr, &client_addr_len ) ) == SOCKET_ERROR ) {
      if( errno == EINTR ) {
        /* a system interrupt occurred -- ignore it and continue waiting */
        continue;
      }
      report_error( "accept", parms );
      return -1;
    } else {
      break;
    }
  }

  return client;
}


static int send_line_raw( int output, SYSTEM_PARAMETERS* parms, char* msg ) {
  if( send( output, msg, strlen( msg ), 0 ) < 0 ) {
    report_error( "send_line_raw 1", parms );
    return -1;
  }

  if( send( output, "\r\n", 2, 0 ) < 0 ) {
    report_error( "send_line_raw 2", parms );
    return -1;
  }

  return 0;
}


static int send_line( int output, SYSTEM_PARAMETERS* parms, char* fmt, ... ) {
  va_list arg;
  char    line[ 1024 ];

  va_start( arg, fmt );
  vsnprintf( line, sizeof( line ), fmt, arg );
  va_end( arg );

  if( send_line_raw( output, parms, line ) != 0 ) {
    return -1;
  }

  return 0;
}


static int handle_connection( int client, int server, SYSTEM_PARAMETERS* parms ) {
  int   remote_server;
  char  buffer[ MAX_BUFFER_LENGTH ];
  char  volatile_buffer[ MAX_BUFFER_LENGTH ];
  char* ptr;
  char* iter;
  char* next;
  int   return_code = -1;
  int   i;
  int   eof;
  int   reading_headers;
  int   tunnel_count;

  DBG_MSG(( "handle_connection\n" ));

  if( ( remote_server = smtp_server_connect( parms ) ) == SOCKET_ERROR ) {
    send_line( client, parms, "421 %s. Proxy SMTP Service temporarily unavailable.", parms->helo_string );
    close( client );
    return -1;
  }

  DBG_MSG(( "reading initial reply from remote server\n" ));

  if( read_reply( remote_server, client, parms ) < 0 ) {
    send_line( client, parms, "421 %s. Proxy SMTP Service temporarily unavailable.", parms->helo_string );
    close( client );
    return -1;
  }

  DBG_MSG(( "entering forward main loop\n" ));

  /* main loop */
  while( 1 ) {
    if( read_line( client, buffer, sizeof( buffer ), &eof, parms ) < 0 ) {
      report_error( "handle_connection->read_line 1", parms );
      break;
    } else if( eof ) {
      DBG_MSG(( "end of stream reached without QUIT command\n" ));
      return_code = 0;
      break;
    }
    i = strlen( buffer );
    while( i > 0 && isspace(buffer[i-1]) ) buffer[--i] = 0;
    strcpy( volatile_buffer, buffer );

    DBG_MSG(( "interpreting '%s'\n", buffer ));
    if( ( ptr = strtok_r( volatile_buffer, " ", &next ) ) == NULL ) {
      DBG_MSG(( "no command entered\n" ));
      send_line( client, parms, "500 Syntax error - No command entered." );

    } else if( strcasecmp( ptr, "QUIT" ) == 0 ) {
      DBG_MSG(( "quit received\n" ));
      send_line( remote_server, parms, "QUIT" );
      send_line( client, parms, "221 %s. SMTP Proxy connection sucessfully closed.", parms->helo_string );
      return_code = 0;
      break;

    } else if( strcasecmp( ptr, "HELO" ) == 0 || strcasecmp( ptr, "EHLO" ) == 0 ) {
      DBG_MSG(( "hello received\n" ));
      send_line( remote_server, parms, "%s %s", ptr, parms->helo_string );
      if( read_reply( remote_server, client, parms ) < 0 ) {
        break;
      }

    } else if( strcasecmp( ptr, "DATA" ) == 0 ) {
      DBG_MSG(( "data received (%s)\n", buffer ));
      send_line_raw( remote_server, parms, buffer );
      if( read_line( remote_server, buffer, sizeof( buffer ), &eof, parms ) < 0 ) {
        report_error( "DATA read_line", parms );
        break;
      }

      send_line_raw( client, parms, buffer );
      if( strlen( buffer ) > 3 && ( buffer[0] == '4' || buffer[0] == '5' ) ) {
        continue;
      }

      reading_headers = 1;
      tunnel_count = 0;
      do {
        if( read_line( client, buffer, sizeof( buffer ), &eof, parms ) < 0 ) {
          report_error( "DATA read_line 2", parms );
          break;
        }

        if( reading_headers ) {
          if( buffer[0] == 0 ) {
            reading_headers = 0;
            tunnel_count++;
            send_line( remote_server, parms, "X-smtpproxy-%02d: %s", tunnel_count, parms->helo_string );
          } else if( strncasecmp( buffer, "X-smtpproxy-", 12 ) == 0 ) {
            tunnel_count++;
          }
        }

        DBG_MSG(( "data received (%s)\n", buffer ));
        send_line_raw( remote_server, parms, buffer );
      } while( strcmp( buffer, "." ) != 0 );
      if( read_reply( remote_server, client, parms ) < 0 ) {
        break;
      }

    } else if( strcasecmp( ptr, "VERB" ) == 0 ) {
      DBG_MSG(( "verb received\n" ));
      send_line( client, parms, "500 Gatewaying of VERB command is not implemented." );

    } else {
      DBG_MSG(( "default received\n" ));
      send_line_raw( remote_server, parms, buffer );
      if( read_reply( remote_server, client, parms ) < 0 ) {
        break;
      }
    }
  }

  close( remote_server );
  close( client );

  VERBOSE( parms, 1, log_message( parms, "child connection closed" ) );
  DBG_MSG(( "returning with a %d\n", return_code ));
  return return_code;
}


static void cleanup( SYSTEM_PARAMETERS* parms ) {
  if( parms->log_file != NULL ) {
    fclose( parms->log_file );
  }
}


int main( int argc, char* argv[] ) {
  int               c;
  int               server;
  int               client;
  pid_t             child;
  SYSTEM_PARAMETERS sys_parms;

  memset( &sys_parms, 0, sizeof( sys_parms ) );
  sys_parms.local_port = 25;
  sys_parms.remote_port = 25;
  sys_parms.max_clients = 5;

  gethostname( sys_parms.helo_string, sizeof( sys_parms.helo_string ) );
  DBG_MSG(( "got hostname: '%s'\n", sys_parms.helo_string ));

  while( ( c = getopt( argc, argv, "p:r:d:m:s:l:vh" ) ) != -1 ) {
    switch( c ) {
      case 'p':
        sys_parms.local_port = atoi( optarg );
        break;
      case 'r':
        sys_parms.remote_port = atoi( optarg );
        break;
      case 'm':
        sys_parms.max_clients = atoi( optarg );
        break;
      case 'd':
        strcpy( sys_parms.remote_address, optarg );
        break;
      case 's':
        strcpy( sys_parms.helo_string, optarg );
        break;
      case 'l':
        if( sys_parms.log_file != NULL ) {
          fclose( sys_parms.log_file );
        }
        sys_parms.log_file = fopen( optarg, "w" );
        break;
      case 'v':
        sys_parms.verbose_level++;
        break;
      case 'h':
        usage();
        cleanup( &sys_parms );
        return 0;
      default:
        usage();
        cleanup( &sys_parms );
        return 0;
    }
  }

  if( sys_parms.local_port == 0 ) {
    log_message( &sys_parms, "You must specify a local port for the proxy to listen on.\n" );
    cleanup( &sys_parms );
    return 0;
  }
  if( sys_parms.remote_port == 0 ) {
    log_message( &sys_parms, "You must specify a local port for the proxy to listen on.\n" );
    cleanup( &sys_parms );
    return 0;
  }
  if( sys_parms.max_clients <= 0 ) {
    log_message( &sys_parms, "You must allow at least one client to run.\n" );
    cleanup( &sys_parms );
    return 0;
  }
  if( sys_parms.remote_address[0] == 0 ) {
    log_message( &sys_parms, "You must specify the address of a SMTP server or proxy to relay to.\n" );
    cleanup( &sys_parms );
    return 0;
  }

  DBG_MSG(( "successfully parsed parameters -- everything looks good (p:%d, r:%d, m:%d, d:%s, s:%s)\n",
            sys_parms.local_port, sys_parms.remote_port, sys_parms.max_clients, sys_parms.remote_address,
            sys_parms.helo_string ));

  DBG_MSG(( "calling server_socket(...)\n" ));
  server = server_socket( &sys_parms );

  if( server < 0 ) {
    DBG_MSG(( "server_socket failed -- quitting\n" ));
    cleanup( &sys_parms );
    return 0;
  }

  DBG_MSG(( "setting child signal handler\n" ));
  signal( SIGCHLD, &child_signal );

  DBG_MSG(( "beginning accept loop\n" ));
  VERBOSE( &sys_parms, 1, log_message( &sys_parms, "server is now waiting for connections..." ) );

  g_active_clients = 0;
  while( 1 ) {
    DBG_MSG(( "calling server_accept\n" ));
    client = server_accept( server, &sys_parms );
    if( client == SOCKET_ERROR ) {
      DBG_MSG(( "client accept failed (%d:%s)", errno, strerror( errno ) ));
      continue;
    }

    if( g_active_clients >= sys_parms.max_clients ) {
      log_message( &sys_parms, "max clients (%d) reached, new connection dropped", sys_parms.max_clients );
      send_line( client, &sys_parms, "421 %s. Maximum proxy clients (%d) reached -- please try again.",
                 sys_parms.helo_string, sys_parms.max_clients );
      close( client );
      continue;
    }

    DBG_MSG(( "forking\n" ));
    child = fork();
    if( child == -1 ) {
      report_error( "fork", &sys_parms );
      send_line( client, &sys_parms, "421 %s. Error forking process.",
                 sys_parms.helo_string, sys_parms.max_clients );
      close( client );
      continue;
    }

    if( child == 0 ) { /* inside the child process */
      VERBOSE( &sys_parms, 1, log_message( &sys_parms, "child connection received" ) );
      handle_connection( client, server, &sys_parms );
      exit( 0 );
    } else {
      g_active_clients++;
    }

    close( client );
  }

  DBG_MSG(( "closing server socket\n" ));
  close( server );

  cleanup( &sys_parms );

  return 0;
}

